<?php

namespace Drupal\cookie_bot_protection;

use Drupal\cookie_bot_protection\Form\SettingsForm;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Site\Settings;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Symfony\Component\HttpFoundation\Cookie;
use Symfony\Component\HttpFoundation\IpUtils;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\HttpKernelInterface;

/**
 * Cookie Bot Protection Middleware object.
 */
class CookieBotProtectionMiddleware implements HttpKernelInterface {

  use StringTranslationTrait;

  /**
   * The decorated kernel.
   *
   * @var \Symfony\Component\HttpKernel\HttpKernelInterface
   */
  protected HttpKernelInterface $httpKernel;

  /**
   * The configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected ImmutableConfig $config;

  /**
   * Constructs a RequestEventSubscriber object.
   *
   * @param \Symfony\Component\HttpKernel\HttpKernelInterface $http_kernel
   *   The following kernel middleware to call.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The service to get the configurations.
   */
  public function __construct(HttpKernelInterface $http_kernel, ConfigFactoryInterface $config_factory) {
    $this->httpKernel = $http_kernel;
    $this->config = $config_factory->get(SettingsForm::SETTINGS);
  }

  /**
   * {@inheritdoc}
   */
  public function handle(Request $request, int $type = self::MAIN_REQUEST, bool $catch = TRUE): Response {
    if ($type === self::SUB_REQUEST || !$this->isNeeded($request)) {
      return $this->httpKernel->handle($request, $type, $catch);
    }

    $hash_headers = [
      'accept_language' => $request->getLanguages(),
      'user_agent' => $request->headers->get('User-Agent'),
      'ip' => $request->getClientIp(),
    ];
    $hash_payload = json_encode($hash_headers);

    $salt = (string) Settings::get('hash_salt');
    $challenge_auth = hash_hmac('sha256', $hash_payload . 'checked', $salt);
    $challenge = hash_hmac('sha256', $hash_payload . 'unchecked', $salt);
    $request_challenge = $request->cookies->get('drupal_cbp_challenge');
    $challenge_param = $request->query->get('drupal_cbp_check');

    $response = NULL;
    if ($challenge_param == "1") {
      if ($request_challenge === $challenge) {
        $response = $this->authorize($request, $challenge_auth);
      }
      elseif ($request_challenge === $challenge_auth) {
        $request->query->remove('drupal_cbp_check');
        $request->overrideGlobals();
        $response = new RedirectResponse($request->getUri());
      }
      else {
        $response = $this->deny($request);
      }
    }
    elseif ($request_challenge !== $challenge_auth) {
      $response = $this->challenge($request, $challenge);
    }

    if ($response !== NULL) {
      if ($this->config->get('header_response_debug')) {
        $response->headers->add([
          'Drupal-CBP-Debug' => 'TRUE',
          'Drupal-CBP-Date' => (new DrupalDateTime())->format('c.v'),
          'Drupal-CBP-Request-Challenge' => $request_challenge,
          'Drupal-CBP-Challenge' => $challenge,
          'Drupal-CBP-Checked-Challenge' => $challenge_auth,
          'Drupal-CBP-Checked-Query' => $challenge_param,
          'Drupal-CBP-Request-Headers' => $hash_payload,
        ]);
      }
      return $this->prepareResponse($response);
    }

    return $this->httpKernel->handle($request, $type, $catch);
  }

  /**
   * Check if the challenge need to be performed.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return bool
   *   The result of the validation.
   *
   * @throws \Exception
   */
  protected function isNeeded(Request $request): bool {

    if (!$this->isNeededUrls($request)) {
      return FALSE;
    }

    if ($this->isUaWhitelisted($request)) {
      return FALSE;
    }

    if ($this->isIpWhitelisted($request)) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Check if the current url is covered by the cookie protection.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return bool
   *   Is the current url is part of the protected page list.
   */
  protected function isNeededUrls(Request $request): bool {
    $url_patterns = $this->config->get('url_protected_patterns');
    if (empty($url_patterns)) {
      return FALSE;
    }
    $uri = $request->getRequestUri();
    $pattern_found = FALSE;
    foreach ($url_patterns as $url_pattern) {
      $pcre_pattern = preg_replace('/\\\\u([a-fA-F0-9]{4})/', '\\x{\\1}', $url_pattern);
      $matches = [];
      if ($pattern_found |= (preg_match('{' . $pcre_pattern . '}u', $uri, $matches) > 0)) {
        break;
      }
    }

    return $pattern_found;
  }

  /**
   * Check if the User-Agent is whitelisted.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return bool
   *   Whitelist result.
   */
  protected function isUaWhitelisted(Request $request): bool {
    $ua_whitelist = $this->config->get('ua_whitelist_patterns');
    if (empty($ua_whitelist)) {
      return FALSE;
    }

    $user_agent = $request->headers->get('User-Agent');
    $pattern_found = FALSE;
    foreach ($ua_whitelist as $url_pattern) {
      $pcre_pattern = preg_replace('/\\\\u([a-fA-F0-9]{4})/', '\\x{\\1}', $url_pattern);
      $matches = [];
      if ($pattern_found |= (preg_match('{' . $pcre_pattern . '}u', $user_agent, $matches) > 0)) {
        break;
      }
    }

    return $pattern_found;
  }

  /**
   * Check if the Ip is whitelisted.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return bool
   *   Whitelist result.
   */
  protected function isIpWhitelisted(Request $request): bool {
    $ip_whitelist = $this->config->get('ip_whitelist');
    if (empty($ip_whitelist)) {
      return FALSE;
    }

    return IpUtils::checkIp($request->getClientIp(), $ip_whitelist);
  }

  /**
   * Prepare the redirect to the challenge and set the corresponding cookie.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   * @param string $challenge
   *   The redirect challenge to set to a cookie.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The challenge response.
   */
  protected function challenge(Request $request, string $challenge): Response {
    $request->query->set('drupal_cbp_check', '1');
    $request->overrideGlobals();
    $cookie = Cookie::create('drupal_cbp_challenge', $challenge);
    $response = new RedirectResponse($request->getUri());
    $response->headers->setCookie($cookie);
    return $response;
  }

  /**
   * Return the authorize response and set cookie.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   * @param string $challenge
   *   The authorization to set to a cookie.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The authorize response.
   */
  protected function authorize(Request $request, string $challenge): Response {
    $request->query->remove('drupal_cbp_check');
    $request->overrideGlobals();
    $response = new RedirectResponse($request->getUri());
    $cookie = Cookie::create('drupal_cbp_challenge', $challenge, 0, NULL, NULL, NULL, TRUE);
    $response->headers->setCookie($cookie);
    return $response;
  }

  /**
   * Return a deny response.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The current request.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The modified response.
   */
  protected function deny(Request $request): Response {
    $redirect_error_delay = (int) $this->config->get('redirect_error_delay');
    $response = new Response();
    $response->setStatusCode(401);

    /*
     * Allow retry if eventually the challenge was still on a legitimate
     * user request.
     */
    if ($redirect_error_delay > 0) {
      $request->query->remove('drupal_cbp_check');
      $request->overrideGlobals();
      // Refresh behavior need at least one content character to display.
      $response->setContent(" ");
      $response->headers->set('Refresh', sprintf('%d; url=%s', $redirect_error_delay, $request->getUri()));
    }

    $response->headers->removeCookie('drupal_cbp_challenge');
    return $response;
  }

  /**
   * Prepare the redirect response with non-cacheability.
   *
   * @param \Symfony\Component\HttpFoundation\Response $response
   *   The response to prepare.
   *
   * @return \Symfony\Component\HttpFoundation\Response
   *   The modified response.
   */
  protected function prepareResponse(Response $response): Response {
    $response->headers->set('Cache-Control', 'no-cache, no-store, must-revalidate');
    $response->setMaxAge(0);
    $response->setSharedMaxAge(0);
    return $response;
  }

}
